package core

import (
	"fmt"
	"slices"

	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/ssa"
	"git.urbach.dev/cli/q/src/token"
	"git.urbach.dev/cli/q/src/types"
)

// Evaluate converts an expression to an SSA value.
func (f *Function) Evaluate(expr *expression.Expression) (ssa.Value, error) {
	if expr.IsLeaf() {
		switch expr.Token.Kind {
		case token.Identifier:
			name := expr.Token.String(f.File.Bytes)
			value, exists := f.Identifiers[name]

			if !exists {
				function, exists := f.All.Functions[f.File.Package+"."+name]

				if !exists {
					return nil, errors.New(&UnknownIdentifier{Name: name}, f.File, expr.Token.Position)
				}

				f.Dependencies.Add(function)
				v := f.AppendFunction(function.UniqueName, function.Type)
				v.Source = ssa.Source(expr.Source)
				return v, nil
			}

			return value, nil

		case token.Number:
			number, err := f.ToNumber(expr.Token)

			if err != nil {
				return nil, err
			}

			v := f.AppendInt(number)
			v.Source = ssa.Source(expr.Source)
			return v, nil

		case token.String:
			data := expr.Token.Bytes(f.File.Bytes)
			data = Unescape(data)
			v := f.AppendBytes(data)
			v.Source = ssa.Source(expr.Source)
			return v, nil
		}

		return nil, errors.New(InvalidExpression, f.File, expr.Token.Position)
	}

	switch expr.Token.Kind {
	case token.Call:
		children := expr.Children
		isSyscall := false

		if children[0].Token.Kind == token.Identifier {
			funcName := children[0].String(f.File.Bytes)

			if funcName == "len" {
				identifier := children[1].String(f.File.Bytes)
				length, exists := f.Identifiers[identifier+".len"]

				if !exists {
					return nil, errors.New(&UnknownIdentifier{Name: identifier + ".len"}, f.File, expr.Token.Position)
				}

				return length, nil
			}

			if funcName == "syscall" {
				children = children[1:]
				isSyscall = true
			}
		}

		args := make([]ssa.Value, len(children))

		for i, child := range slices.Backward(children) {
			value, err := f.Evaluate(child)

			if err != nil {
				return nil, err
			}

			args[i] = value
		}

		if isSyscall {
			syscall := &ssa.Syscall{
				Arguments: args,
				Source:    ssa.Source(expr.Source),
			}

			return f.Append(syscall), nil
		} else {
			name := args[0].(*ssa.Function).UniqueName
			fn := f.All.Functions[name]
			parameters := args[1:]

			if len(parameters) != len(fn.Input) {
				return nil, errors.New(&ParameterCountMismatch{Function: name, Count: len(parameters), ExpectedCount: len(fn.Input)}, f.File, expr.Source[0].Position)
			}

			for i, param := range parameters {
				if !types.Is(param.Type(), fn.Input[i].Typ) {
					return nil, errors.New(&TypeMismatch{
						Encountered:   param.Type().Name(),
						Expected:      fn.Input[i].Typ.Name(),
						ParameterName: fn.Input[i].Name,
					}, f.File, param.Start())
				}
			}

			call := &ssa.Call{
				Arguments: args,
				Source:    ssa.Source(expr.Source),
			}

			return f.Append(call), nil
		}

	case token.Dot:
		left := expr.Children[0]
		right := expr.Children[1]
		label := fmt.Sprintf("%s.%s", left.String(f.File.Bytes), right.String(f.File.Bytes))
		function, exists := f.All.Functions[label]

		if !exists {
			return nil, errors.New(&UnknownIdentifier{Name: label}, f.File, left.Token.Position)
		}

		f.Dependencies.Add(function)
		v := f.AppendFunction(function.UniqueName, function.Type)
		v.Source = ssa.Source(expr.Source)
		return v, nil
	}

	return nil, nil
}